{% extends '!layout.html' %}

{% block extrahead %}

	<!-- frame blocking -->
	<script type="text/javascript">
		if (top !=self) {
			top.location=self.location;
		}
	</script>



	<!-- NIST header and footer -->
  <link rel="stylesheet" href="https://pages.nist.gov/nist-header-footer/css/nist-combined.css">
  <script src="https://code.jquery.com/jquery-1.12.4.min.js" type="text/javascript"></script>
  <!-- <script src="https://pages.nist.gov/nist-header-footer/js/nist-header-footer.js" type="text/javascript" defer="defer"></script> -->


  <script type="text/javascript" src="https://pages.nist.gov/leaveNotice/js/jquery.leaveNotice-nist.min.js"></script>
  <script>
   $(document).ready(function(){
       // Mark external (non-nist.gov) A tags with class "external"
       //If the adress start with https and ends with nist.gov
       var re_nist = new RegExp('^https?:\/\/((^\/)*\.)*nist\\.gov(\/|$)');
       //Regex to find address that start with https
       var re_absolute_address = new RegExp('^((https?:)?\/\/)');
       $("a").each(function(){
           var url=$(this).attr('href');
           if(re_nist.test(url) || !re_absolute_address.test(url)){
               $(this).addClass('local');
           }else{
               //This a href appears to be external, so tag it
               $(this).addClass('external');
           }
       });
       // Add leaveNotice to external A elements
       $('a.external').leaveNotice();
   });
  </script>
  <link rel="stylesheet" type="text/css" href="https://pages.nist.gov/leaveNotice/css/jquery.leaveNotice.css" />

{% endblock %}
