{% extends '!footer.html' %}

{% block extrafooter %}


	<br></br>
	{% include "boilerplate-footer.html" %}

  <!-- Global site tag (gtag.js) - Google Analytics -->

  <script async src="https://www.googletagmanager.com/gtag/js?id=UA-115121490-6"></script>
  <script>
    window.dataLayer = window.dataLayer || [];
    function gtag(){dataLayer.push(arguments);}
    gtag('js', new Date());

    gtag('config', 'UA-115121490-6');
  </script>

{% endblock %}
